#include <iostream>
#include <FL/gl.H>

using namespace std;

#include "vector_3d.h"
#include "objects.h"
#include "line.h"

Line::Line ()
    : Object (), 
      pta (position), 
      ptb (pta + Vector_3d (10,0,0))
{
}

Line::Line (const Point_3d &a, const Point_3d &b)
    : Object (a, Vector_3d ( 0, 0, 0)),
      pta (a), ptb (b)
{
}

Line::Line (const Line &that)
{
    *this = that;
}

Line &Line::operator= (const Line &that)
{
    if (this == &that)
        return *this;
    this->position = that.position;
    this->velocity = that.velocity;
    this->collided = that.collided;
    this->pta = that.pta;
    this->ptb = that.ptb;
    return *this;
}

Point_3d Line::get_closest_point (Point_3d P, bool segmentClamp)
{ 
    Vector_3d AP = P - pta;
    Vector_3d AB = ptb - pta;

    double ab2 = pow (AB.get_x(), 2) + pow (AB.get_y(), 2);
    double ap_ab = AP.dot_product(AB);

    double t = ap_ab / ab2;
    if (segmentClamp)
    {
        if (t < 0.0f)
            t = 0.0f;
        else if (t > 1.0f) 
            t = 1.0f;
    }
    Point_3d closest = pta + AB * t;
    return closest;
}

bool Line::collide (Object &other)
{
    Point_3d closest = get_closest_point (other.get_pos(), true);
    double distfromline = (closest.distance (other.get_pos())).magnitude();
    //cout << distfromline << endl;
    //cout << get_col() << endl;
    if (distfromline < 25 && !other.get_col ())
    {
        //cout << "collision!" << endl;
        other.set_col (true);
        resolve_collision (other);
        return true;
    }
    else if (distfromline > 30 && other.get_col ())
        other.set_col (false);
    return false;
}

void Line::resolve_collision (Object &other)
{
    //(1) find unit normal & unit tangent
    Point_3d temp = get_closest_point (other.get_pos(), true);
    Vector_3d normvec = other.get_pos() - temp;
    Vector_3d unitnormvec = normvec.unit_vector();

    Vector_3d tangvec = Vector_3d (-normvec.get_y(), normvec.get_x(), 0);
    Vector_3d unittangvec = tangvec.unit_vector();

    //(2) get velocities
    Vector_3d linevel = velocity;
    Vector_3d othervel = other.get_vel();

    //(3) resolve velocities into normal and tangents
    double v1norm = unitnormvec.dot_product (linevel);
    double v1tang = unittangvec.dot_product (linevel);

    double v2norm = unitnormvec.dot_product (othervel);
    double v2tang = unittangvec.dot_product (othervel);

    //(4) find new tangential velocities
    double v1primetang = v1tang;
    double v2primetang = v2tang;

    //(5) find new normal velocities (scalar)
    double v1primenorm = (linevel.magnitude());
    double v2primenorm = (othervel.magnitude());

    //(6) convert (5) to vectors
    Vector_3d v1primenormvec = v1primenorm * unitnormvec;
    Vector_3d v1primetangvec = v1primetang * unittangvec;

    Vector_3d v2primenormvec = v2primenorm * unitnormvec;
    Vector_3d v2primetangvec = v2primetang * unittangvec;

    //(7) add tangent' and normal' to get new velocities
    velocity = Vector_3d (0,0,0); //v1primenormvec + v2primetangvec;
    other.set_vel( v2primenormvec + v2primetangvec );//(v2primenormvec + v2primetangvec) * .85);
}

void Line::update ()
{
    position = position + velocity;
}

void Line::draw ()
{
    Vector_3d offset = ptb - pta;
    glTranslatef( (float) position.get_x(), (float) position.get_y(), 0);
    //glColor3f (1.0f, 0.0f, 1.0f);
    glBegin (GL_LINES);
        glVertex2f ( 0, 0);
        glVertex2f ( (float) offset.get_x(), (float) offset.get_y());
    glEnd();
}

Point_3d Line::get_a () const
{
    return pta;
}

Point_3d Line::get_b () const
{
    return ptb;
}

void Line::set_a (const Point_3d &newa)
{
    pta = newa;
}

void Line::set_b (const Point_3d &newb)
{
    ptb = newb;
}